
import logging
import colorama
import coloredlogs
from .settings import *


CRITICAL = 0 ### logs only fatal errors (not handled)
ERROR = 1 ### logs only errors
WARNING = 2 ### logs also warnings
INFO = 3 ### logs also information
DEBUG = 4 ### logs also debug information
EXCEPTION = 5 ### logs also unhandled exceptions
PACKAGE = 6 ### also sets built-in logger level to logging.INFO
FULL = 7 ### also sets built-in logger level to logging.DEBUG


FIELD_STYLES = dict(
    asctime=dict(color='green', faint=True, bold=BOLD_LOGS),
    levelname=dict(color='cyan', faint=True, bold=BOLD_LOGS),
)

LEVEL_STYLES = dict(
    debug=dict(color='white', faint=True, bold=BOLD_LOGS),
    info=dict(color='white', bold=BOLD_LOGS),
    warning=dict(color='yellow', bold=BOLD_LOGS),
    error=dict(color='red', bold=BOLD_LOGS),
    critical=dict(color='red', bold=BOLD_LOGS)
)

### to control logging verbosity of external libraries
verbosity_map = {
    CRITICAL : logging.CRITICAL,
    ERROR : logging.ERROR,
    WARNING : logging.WARNING,
    INFO : logging.WARNING,
    DEBUG : logging.WARNING,
    EXCEPTION : logging.WARNING,
    PACKAGE : logging.INFO,
    FULL : logging.DEBUG,
}

level_name_map = {
    'CRITICAL': CRITICAL,
    'ERROR': ERROR,
    'WARNING': WARNING,
    'INFO': INFO,
    'DEBUG': DEBUG,
    'EXCEPTION': EXCEPTION,
    'PACKAGE': PACKAGE,
    'FULL': FULL,
}


def get_user_data_dir(appname):
    from pathlib import Path
    import os
    if WIN_PLATFORM:
        import winreg
        key = winreg.OpenKey(
            winreg.HKEY_CURRENT_USER,
            r"Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders"
        )
        dir_,_ = winreg.QueryValueEx(key, "Local AppData")
        result = Path(dir_).resolve(strict=False)
    elif MAC_PLATFORM:
        result = Path('~/Library/Application Support/').expanduser()
    elif NIX_PLATFORM:
        result = Path(os.getenv('XDG_DATA_HOME', "~/.local/share")).expanduser()
    else:
        raise NotImplementedError
    os.makedirs(result.joinpath(appname, 'logs'), exist_ok=True)
    return result.joinpath(appname, 'logs')


def get_log_file_name():
    from datetime import datetime
    filename = datetime.now().strftime('%Y-%m-%d.log')
    return os.path.join(get_user_data_dir('dso'), filename) 


class LoggerClass():

    def get_log_format(self):
        log_format = ''
        if TIMESTAMP_LOGS: log_format += '%(asctime)s'
        if LABEL_LOG_LEVELS: log_format += ' [%(levelname)-8s]'
        log_format += ' %(message)s'
        return log_format

    def __init__(self, level):
        self.level = level
        logging.basicConfig(format=self.get_log_format(), datefmt=LOG_TIME_FORMAT, level=self.mapped_level, force=True)
        self.logger = logging.getLogger('dso')
        self.logger.setLevel(logging.DEBUG)
        #### Add file handler to capture all logs, regardless of verbosity level
        fh = logging.FileHandler(get_log_file_name())
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(logging.Formatter(self.get_log_format(), LOG_TIME_FORMAT))
        self.logger.addHandler(fh)
        ### also add it to the root to capture messages from other libraries as well
        self.logger.root.addHandler(fh)
        if COLORED_LOGS:        
            ### set level always to maximum (as the level is controlled by mapped_level), and attach it to root logger
            coloredlogs.install(logger=logging.root, 
                                level='DEBUG', 
                                fmt=self.get_log_format(),
                                level_styles=LEVEL_STYLES,
                                field_styles=FIELD_STYLES,
                                )
        else:
            ch = logging.StreamHandler()
            ch.setLevel(logging.DEBUG)
            ch.setFormatter(logging.Formatter(self.get_log_format(), LOG_TIME_FORMAT))
            self.logger.addHandler(ch)
            ### also add it to the root to capture messages from other libraries as well
            self.logger.root.addHandler(ch)

    ### maps dso log verbosity to internal verbosity
    def _map_level(self, level):
        return verbosity_map[level]

    @property
    def mapped_level(self):
        return self._map_level(self.level)

    def set_verbosity(self,level):
        self.level = level
        logging.root.setLevel(self.mapped_level)

    def critical(self, msg, force=True):
        self.logger.propagate = force or self.level >= CRITICAL
        self.logger.critical(msg)

    def error(self, msg, force=True):
        self.logger.propagate = force or self.level >= ERROR
        self.logger.error(msg)

    def warn(self, msg, force=False):
        self.logger.propagate = force or self.level >= WARNING
        self.logger.warning(msg)

    def info(self, msg, force=False):
        self.logger.propagate = force or self.level >= INFO
        self.logger.info(msg)

    def debug(self, msg, force=False):
        self.logger.propagate = force or self.level >= DEBUG
        self.logger.debug(msg)

Logger = LoggerClass(level_name_map[DEFAULT_LOG_LEVEL])




import re
import boto3
from .contexts import Contexts, Context
from .logger import Logger
from .stages import Stages
import logging

logging.getLogger('botocore').setLevel(Logger.mapped_level)
logging.getLogger('boto').setLevel(Logger.mapped_level)
logging.getLogger('boto3').setLevel(Logger.mapped_level)

ssm_resource_type = {
    'String': 'parameter',
    'SecureString': 'secret',
    'StringList': 'template',
}

ssm_null_value = '352e9018-684d-4c4f-a7d2-ceeb9e0fe947'
ssm_empty_value = '3b22cef4-8786-48e9-8521-0a1e45891a3c'


### AWS SSM Parameter store does not allow {{}}
def escape_curly_brackets(contents):
    return contents.replace('{{', r'\{\{').replace('}}', r'\}\}')


def unescape_curly_brackets(contents):
    return contents.replace(r'\{\{', '{{').replace(r'\}\}', '}}')


### AWS SSM Parameter store does not allow empty string or null values
def encode_nulls(value):
    if value is None:
        return ssm_null_value
    elif value == '':
        return ssm_empty_value
    else:
        return value


def decode_nulls(value):
    if value == ssm_null_value:
        return None
    elif value == ssm_empty_value:
        return ''
    else:
        return value



def get_ssm_path(context, key=None, path_prefix=''):
    return path_prefix + context.get_path(key)



def load_ssm_path(result, config, path, parameter_type, path_prefix='', decrypt=False, filter=None):
    ssm = boto3.session.Session().client(service_name='ssm')
    p = ssm.get_paginator('get_parameters_by_path')
    paginator = p.paginate(Path=path, Recursive=True, WithDecryption=decrypt, ParameterFilters=[{'Key': 'Type', 'Values': [parameter_type]}]).build_full_result()
    for parameter in paginator['Parameters']:
        key = parameter['Name'][len(path)+1:]
        if filter and not re.match(filter, key): continue
        if key in result:
            Logger.warn("Inherited {0} '{1}' has been overridden.".format(ssm_resource_type[parameter_type] , key))
        ctx_path = path[len(path_prefix):]
        ctx = Context(*Contexts.parse_path(ctx_path)[0:4])
        details = {
                    'Value': unescape_curly_brackets(parameter['Value']) if parameter_type == 'StringList' else decode_nulls(parameter['Value']), 
                    'Path': parameter['Name'],
                    'Version': parameter['Version'],
                    'Stage': ctx.short_stage,
                    'Scope': ctx.scope_translation,
                    'Origin': {
                        'Namespace': ctx.namespace,
                        'Project': ctx.project,
                        'Application': ctx.application,
                        'Stage': ctx.stage,
                    },
                    'Date': parameter['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                    'ARN': parameter['ARN'],
                }
        result[key] = details
    return result



def load_context_ssm_parameters(config, parameter_type, path_prefix='', decrypt=False, uninherited=False, filter=None):
    ### construct search path in hierachy with no key specified
    paths = Contexts.get_hierachy_paths(context=config.context, key=None, path_prefix=path_prefix, ignore_stage=config.stage is None, uninherited=uninherited)
    parameters = {}
    for path in paths:
        Logger.debug(f"Loading SSM parameters: path={path}")
        load_ssm_path(config=config, result=parameters, path=path, parameter_type=parameter_type, path_prefix=path_prefix, decrypt=decrypt, filter=filter)
    return parameters



def locate_ssm_parameter_in_context_hierachy(config, key, path_prefix='', uninherited=False):
    result = {}
    paths = Contexts.get_hierachy_paths(context=config.context, key=key, path_prefix=path_prefix, ignore_stage=config.stage is None, uninherited=uninherited, reverse=True)
    ssm = boto3.session.Session().client(service_name='ssm')
    for path in paths:
        Logger.debug(f"Describing SSM parameters: path={path}")
        response = ssm.describe_parameters(ParameterFilters=[{'Key':'Name', 'Values':[path]}])
        if len(response['Parameters']) > 0:
            ctx_path = path[len(path_prefix):]
            ctx = Context(*Contexts.parse_path(ctx_path)[0:4])
            result = response['Parameters'][0]
            result['Stage'] = ctx.short_stage
            result['Scope'] = ctx.scope_translation
            result['Origin']= {
                'Namespace': ctx.namespace,
                'Project': ctx.project,
                'Application': ctx.application,
                'Stage': ctx.stage,
            }
            break
    return result



def assert_ssm_parameter_no_namespace_overwrites(config, key, path_prefix=''):
    """
        check if a parameter will overwrite parent or childern parameters (with the same namespaces) in the same context (always uninherited)
        e.g.: 
            parameter a.b.c would overwrite a.b (super namespace)
            parameter a.b would overwrite all a.b.x (sub namespaces)
    """
    ssm = boto3.session.Session().client(service_name='ssm')
    
    ### check children parameters
    path = path_prefix + config.context.get_path(key)
    response = ssm.describe_parameters(ParameterFilters=[{'Key':'Name', 'Option': 'BeginsWith', 'Values':[f"{path}."]}])
    if len(response['Parameters']) > 0:
        raise DSOException("Parameter key '{0}' is not allowed in the given context becasue it would overwrite '{0}.{1}' and all other parameters in '{0}.*' namespace if any.".format(key,response['Parameters'][0]['Name'][len(path)+1:]))

    ### check parent parameters
    namespaces = key.split('.')
    for n in range(len(namespaces)-1):
        subKey = '.'.join(namespaces[0:n+1])
        path = get_ssm_path(config.context, subKey, path_prefix)
        Logger.debug(f"Describing SSM parameters: path={path}")
        # parameters = ssm.describe_parameters(ParameterFilters=[{'Key':'Type', 'Values':['String']},{'Key':'Name', 'Values':[path]}])
        response = ssm.describe_parameters(ParameterFilters=[{'Key':'Name', 'Values':[path]}])
        if len(response['Parameters']) > 0:
            raise DSOException("Parameter key '{0}' is not allowed in the given context becasue it would overwrite parameter '{1}'.".format(key, subKey))



def put_ssm_paramater(path, value):
    ssm = boto3.session.Session().client(service_name='ssm')
    return ssm.put_parameter(Name=path, Value=encode_nulls(value), Type='String', Overwrite=True)



def put_ssm_secret(path, value):
    ssm = boto3.session.Session().client(service_name='ssm')
    return ssm.put_parameter(Name=path, Value=encode_nulls(value), Type='SecureString', Overwrite=True)



def put_ssm_template(path, contents):
    ssm = boto3.session.Session().client(service_name='ssm')
    return ssm.put_parameter(Name=path, Value=escape_curly_brackets(contents), Type='StringList', Overwrite=True)


def __get_ssm_parameter_history(name, decrypt=False):
    ssm = boto3.session.Session().client(service_name='ssm')
    return ssm.get_parameter_history(Name=name, WithDecryption=decrypt)



def get_ssm_parameter_history(name):
    response = __get_ssm_parameter_history(name)
    for i in range(0, len(response['Parameters'])):
        response['Parameters'][i]['Value'] = decode_nulls(response['Parameters'][i]['Value'])
    return response



def get_ssm_secret_history(name, decrypt=False):
    response = __get_ssm_parameter_history(name, decrypt)
    if decrypt:
        for i in range(0, len(response['Parameters'])):
            response['Parameters'][i]['Value'] = decode_nulls(response['Parameters'][i]['Value'])
    return response



def get_ssm_template_history(name):
    response = __get_ssm_parameter_history(name)
    for i in range(0, len(response['Parameters'])):
        response['Parameters'][i]['Value'] = unescape_curly_brackets(response['Parameters'][i]['Value'])
    return response



def delete_ssm_parameter(name):
    ssm = boto3.session.Session().client(service_name='ssm')
    ssm.delete_parameter(Name=name)




import os
import re
import jinja2
from jinja2 import meta
from .constants import *
from .providers import StoreProvider, Providers
from .parameters import Parameters
from .secrets import Secrets
from .logger import Logger
from .dict_utils import merge_dicts, deflatten_dict
from .exceptions import DSOException
from .stages import Stages
from .config import ContextSource


REGEX_PATTERN = r"^[a-zA-Z]([./a-zA-Z0-9_-]*[a-zA-Z0-9])?$"

class TemplateProvider(StoreProvider):
    def list(self, config, uninherited=False, include_contents=False, filter=None):
        raise NotImplementedError()
    def add(self, config, key, contents, render_path):
        raise NotImplementedError()
    def get(self, config, key, revision=None):
        raise NotImplementedError()
    def history(self, config, key, include_contents=False):
        raise NotImplementedError()
    def delete(self, config, key):
        raise NotImplementedError()


class TemplateService():
    
    @property
    def default_render_path(self):
        return self.config.working_dir

    def validate_key(self, key):
        Logger.info(f"Validating template key '{key}'...")
        if not key:
            raise DSOException(MESSAGES['KeyNull'])
        if key == 'dso' or key.startswith('dso.'):
            raise DSOException(MESSAGES['DSOReserverdKey'].format(key))
        if not re.match(REGEX_PATTERN, key):
            raise DSOException(MESSAGES['InvalidKeyPattern'].format(key, REGEX_PATTERN))
        ### the regex does not check adjacent special chars
        if '..' in key:
            raise DSOException(MESSAGES['InvalidKeyStr'].format(key, '..'))

        if '//' in key:
            raise DSOException(MESSAGES['InvalidKeyStr'].format(key, '//'))

    def get_template_render_path(self, key):
        result = self.config.get_template_render_paths(key)
        if result:
            return result[key]

        return f'.{os.sep}' + os.path.relpath(os.path.join(Templates.default_render_path, key), self.config.working_dir) 

    def list(self, config, uninherited=False, include_contents=False, filter=None):
        self.config = config
        provider = Providers.TemplateProvider()
        Logger.info(f"Start listing templates: namespace={config.get_namespace(ContextSource.Target)}, project={config.get_project(ContextSource.Target)} application={config.get_application(ContextSource.Target)}, stage={config.get_stage(ContextSource.Target, short=True)}, scope={config.scope}")
        response = provider.list(config, uninherited, include_contents, filter)
        for template in response['Templates']:
            key = template['Key']
            template['RenderPath'] = self.get_template_render_path(key)
        
        return response

    def add(self, config, key, contents, render_path):
        self.validate_key(key)
        self.config = config
        provider = Providers.TemplateProvider()
        Logger.info(f"Start adding template '{key}': namespace={config.get_namespace(ContextSource.Target)}, project={config.get_project(ContextSource.Target)} application={config.get_application(ContextSource.Target)}, stage={config.get_stage(ContextSource.Target, short=True)}, scope={config.scope}")
        Logger.debug(f"Template: key={key}, render_path={render_path}")
        result = provider.add(config, key, contents)
        result['RenderPath'] = render_path
        if os.path.abspath(render_path) == os.path.abspath(os.path.join(self.default_render_path, key)):
            config.unregister_template_custom_render_path(key)
        else:
            config.register_template_custom_render_path(key, render_path)
        return result

    def get(self, config, key, revision=None):
        self.config = config
        provider = Providers.TemplateProvider()
        Logger.info(f"Start getting template '{key}': namespace={config.get_namespace(ContextSource.Target)}, project={config.get_project(ContextSource.Target)} application={config.get_application(ContextSource.Target)}, stage={config.get_stage(ContextSource.Target, short=True)}, scope={config.scope}")
        result = provider.get(config, key, revision)
        result['RenderPath'] = self.get_template_render_path(key)
        return result

    def history(self, config, key, include_contents=False):
        self.config = config
        provider = Providers.TemplateProvider()
        Logger.info(f"Start getting the history of template '{key}': namespace={config.get_namespace(ContextSource.Target)}, project={config.get_project(ContextSource.Target)} application={config.get_application(ContextSource.Target)}, stage={config.get_stage(ContextSource.Target, short=True)}, scope={config.scope}")
        return provider.history(config, key, include_contents)

    def delete(self, config, key):
        self.config = config
        provider = Providers.TemplateProvider()
        Logger.info(f"Start deleting template '{key}': namespace={config.get_namespace(ContextSource.Target)}, project={config.get_project(ContextSource.Target)} application={config.get_application(ContextSource.Target)}, stage={config.get_stage(ContextSource.Target, short=True)}, scope={config.scope}")
        result = provider.delete(config, key)
        config.unregister_template_custom_render_path(key)
        return result

    def render(self, config, filter=None):
        self.config = config

        Logger.info(f"Start rendering templates: namespace={config.get_namespace(ContextSource.Target)}, project={config.get_project(ContextSource.Target)} application={config.get_application(ContextSource.Target)}, stage={config.get_stage(ContextSource.Target, short=True)}, scope={config.scope}")

        Logger.info("Loading secrets...")
        secrets = Secrets.list(config, uninherited=False, decrypt=True)

        Logger.info("Loading parameters...")
        parameters = Parameters.list(config, uninherited=False)

        Logger.info("Merging all parameters...")
        merged = deflatten_dict({x['Key']: x['Value'] for x in secrets['Secrets']})
        merge_dicts(deflatten_dict({x['Key']: x['Value'] for x in parameters['Parameters']}), merged)
        merge_dicts(config.meta_vars, merged)

        Logger.info("Loading templates...")
        templates = self.list(config, filter=filter)['Templates']

        loader = jinja2.FileSystemLoader(config.working_dir)
        jinja_env = jinja2.Environment(loader=loader, undefined=jinja2.StrictUndefined)

        rendered = []
        if len(templates) == 0:
            Logger.warn("No template found to render.")
            return rendered

        Logger.info("Rendering templates...")
        for template in templates:
            key = template['Key']

            renderPath = template['RenderPath']
            if os.path.isdir(renderPath):
                raise DSOException("There is an existing directory at the template render path '{renderPath}'.")
            if os.path.dirname(renderPath):
                os.makedirs(os.path.dirname(renderPath), exist_ok=True)

            try:
                jinjaTemplate = jinja_env.from_string(self.get(config, key)['Contents'])
            except:
                Logger.error(f"Failed to load template: {key}")
                raise
            # undeclaredParams = jinja2.meta.find_undeclared_variables(env.parse(template))
            # if len(undeclaredParams) > 0:
            #     Logger.warn(f"Undecalared parameter(s) found:\n{set(undeclaredParams)}")
            try:
                Logger.debug(f"Rendering template: key={key}, render_path={renderPath}")
                if len(loader.searchpath) > 1: loader.searchpath.pop(-1)
                loader.searchpath.append(os.path.dirname(os.path.join(config.working_dir, renderPath)))
                renderedContent = jinjaTemplate.render(merged)
            
            except Exception as e:
                Logger.error(f"Failed to render template: {key}")
                msg = getattr(e, 'message', getattr(e, 'msg', str(e)))
                raise DSOException(msg)

            with open(renderPath, 'w', encoding='utf-8') as f:
                f.write(renderedContent)
            
            rendered.append({
                        'Key':key, 
                        'Scope': template['Scope'],
                        # 'Origin': template['Origin'],
                        'RenderPath': renderPath,
                        })

        return rendered


Templates = TemplateService()


import os
import sys
import signal
import contextlib
from subprocess import Popen, PIPE
from .logger import Logger
from .settings import *
from .file_utils import *
from .settings import *


def get_pager():
    if WIN_PLATFORM:
        return WindowsPager()
    else:
        return PosixPager()


@contextlib.contextmanager
def ignore_ctrl_c():
    original = signal.signal(signal.SIGINT, signal.SIG_IGN)
    try:
        yield
    finally:
        signal.signal(signal.SIGINT, original)


class PagerBase(object):

    def __init__(self, output_stream=sys.stdout):
        self.output_stream = output_stream

    @property
    def PAGER(self):
        return None

    def _get_pager_cmdline(self):
        pager = PAGER or self.PAGER
        return pager.split(' ')


    def _check_pager_cmdline(self):
        pass

    def page(self, contents):
        if USE_PAGER:
            converted_content = self._convert_doc_content(contents)
            self._send_output_to_pager(converted_content)
        else:
            print(contents, flush=True)

    def _send_output_to_pager(self, output):
        self._check_pager_cmdline()
        p = self._popen(self.cmdline, stdin=PIPE)
        p.communicate(input=output)

    def _popen(self, *args, **kwargs):
        return Popen(*args, **kwargs)

    def _convert_doc_content(self, contents: str):
        return contents.encode(encoding='utf-8')


class PosixPager(PagerBase):

    @property
    def PAGER(self):
        return 'less -R'

    def _send_output_to_pager(self, output):
        self._check_pager_cmdline()
        with ignore_ctrl_c():
            p = self._popen(self.cmdline, stdin=PIPE)
            p.communicate(input=output)

    def _check_pager_cmdline(self):
        self.cmdline = self._get_pager_cmdline()
        if not exists_on_path(self.cmdline[0]):
            raise DSOException(f"Pager '{self.cmdline[0]}' not found in PATH.")

class WindowsPager(PagerBase):

    @property
    def PAGER(self):
        if exists_on_path("less.exe"):
            return 'less -R'
        else:
            return 'more'

    def _popen(self, *args, **kwargs):
        kwargs['shell'] = True
        return Popen(*args, **kwargs)


    def _check_pager_cmdline(self):
        self.cmdline = self._get_pager_cmdline()
        if not (exists_on_path(f"{self.cmdline[0]}.exe") or exists_on_path(f"{self.cmdline[0]}.com")):
            raise DSOException(f"Pager '{self.cmdline[0]}' not found in PATH.")

Pager = get_pager()



import boto3
from dsocli.exceptions import DSOException
from dsocli.logger import Logger
from dsocli.providers import Providers
from dsocli.templates import TemplateProvider
from dsocli.stages import Stages
from dsocli.constants import *
from dsocli.dict_utils import set_dict_value
from dsocli.contexts import Contexts
from dsocli.aws_utils import *
from dsocli.settings import *


__default_spec = {
    'path': '/dso/v1/templates',
    # 'namnespace': 'default',
}

def get_default_spec():
    return __default_spec.copy()


class AwsSsmTemplateProvider(TemplateProvider):


    def __init__(self):
        super().__init__('template/aws/ssm/v1')


    def get_path_prefix(self):
        return self.config.template_spec('path')


    def list(self, config, uninherited=False, include_contents=False, filter=None):
        self.config = config
        templates = load_context_ssm_parameters(config=config, parameter_type='StringList', path_prefix=self.get_path_prefix(), uninherited=uninherited, filter=filter)
        result = {'Templates': []}
        for key, details in templates.items():
            item = {
                'Key': key,
                'RevisionId': str(details['Version']),
            }
            item.update(details)
            if include_contents: item['Contents'] = item['Value']
            item.pop('Value')
            result['Templates'].append(item)
        return result


    def add(self, config, key, contents, render_path=None):
        self.config = config
        if len(contents) > 4096:
            raise DSOException(f"This template provider does not support templates larger than 4KB.")
        if not Stages.is_default(config.stage) and not ALLOW_STAGE_TEMPLATES:
            raise DSOException(f"Templates may not be added to stage scopes, as the feature is currently disabled. It may be enabled by adding 'ALLOW_STAGE_TEMPLATES=yes' to the DSO global settings, or adding environment variable 'DSO_ALLOW_STAGE_TEMPLATES=yes'.")
        Logger.debug(f"Checking SSM template '{key}' overwrites: namespace:{config.namespace}, project={config.project}, application={config.application}, stage={config.stage}")
        assert_ssm_parameter_no_namespace_overwrites(config=config, key=key, path_prefix=self.get_path_prefix())
        Logger.debug(f"Locating SSM template '{key}': namespace:{config.namespace}, project={config.project}, application={config.application}, stage={config.stage}")
        found = locate_ssm_parameter_in_context_hierachy(config=config, key=key, path_prefix=self.get_path_prefix(), uninherited=True)
        if found and not found['Type'] == 'StringList':
            raise DSOException(f"Failed to add template '{key}' becasue becasue the key is not available in the given context: namespace:{config.namespace}, project={config.project}, application={config.application}, stage={config.short_stage}")
        path = get_ssm_path(context=config.context, key=key, path_prefix=self.get_path_prefix())
        Logger.debug(f"Adding SSM template: path={path}")
        response = put_ssm_template(path, contents)
        result = {
                'RevisionId': str(response['Version']),
                'Key': key,
                'Stage': config.short_stage,
                'Scope': config.context.scope_translation,
                'Origin': {
                    'Namespace': config.namespace,
                    'Project': config.project,
                    'Application': config.application,
                    'Stage': config.stage,
                },
                'Path': path,
            }
        result.update(response)
        return result


    def get(self, config, key, revision=None):
        self.config = config
        Logger.debug(f"Locating SSM template '{key}': namespace:{config.namespace}, project={config.project}, application={config.application}, stage={config.stage}")
        found = locate_ssm_parameter_in_context_hierachy(config=config, key=key, path_prefix=self.get_path_prefix())
        if not found:
            raise DSOException(f"Template '{key}' not found nor inherited in the given context: namespace:{config.namespace}, project={config.project}, application={config.application}, stage={config.short_stage}")
        else:
            if not found['Type'] == 'StringList':
                raise DSOException(f"Template '{key}' not found in the given context: namespace:{config.namespace}, project={config.project}, application={config.application}, stage={config.short_stage}")
        Logger.debug(f"Getting SSM template: path={found['Name']}")
        response = get_ssm_template_history(found['Name'])
        templates = sorted(response['Parameters'], key=lambda x: int(x['Version']), reverse=True)
        if revision is None:
            ### get the latest revision
            result = {
                    'RevisionId': str(templates[0]['Version']),
                    'Date': templates[0]['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                    'Key': key, 
                    'Scope': found['Scope'],
                    'Origin': found['Origin'],
                    'User': templates[0]['LastModifiedUser'],
                    'Path': found['Name'],
                    'Contents': templates[0]['Value'],
                    }
        else:
            ### get specific revision
            templates = [x for x in templates if str(x['Version']) == revision]
            if not templates:
                raise DSOException(f"Revision '{revision}' not found for template '{key}' in the given context: namespace:{config.namespace}, project={config.project}, application={config.application}, stage={config.short_stage}")
            result = {
                    'RevisionId': str(templates[0]['Version']),
                    'Date': templates[0]['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                    'Key': key, 
                    'Scope': found['Scope'],
                    'Origin': found['Origin'],
                    'Path': found['Name'],
                    'User': templates[0]['LastModifiedUser'],
                    'Contents': templates[0]['Value'],
                    }

        return result



    def history(self, config, key, include_contents=False):
        self.config = config
        Logger.debug(f"Locating SSM template '{key}': namespace:{config.namespace}, project={config.project}, application={config.application}, stage={config.stage}")
        found = locate_ssm_parameter_in_context_hierachy(config=config, key=key, path_prefix=self.get_path_prefix())
        if not found:
            raise DSOException(f"Template '{key}' not found in the given context: namespace:{config.namespace}, project={config.project}, application={config.application}, stage={config.short_stage}")
        else:
            if not found['Type'] == 'StringList':
                raise DSOException(f"Template '{key}' not found in the given context: namespace:{config.namespace}, project={config.project}, application={config.application}, stage={config.short_stage}")
        Logger.debug(f"Getting SSM template: path={found['Name']}")
        response = get_ssm_parameter_history(found['Name'])
        templates = sorted(response['Parameters'], key=lambda x: int(x['Version']), reverse=True)
        if include_contents:
            result = { "Revisions":
                [{
                    'RevisionId': str(template['Version']),
                    'Date': template['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                    'Key': key,
                    # 'Scope': found['Scope'],
                    # 'Origin': found['Origin'],
                    'User': template['LastModifiedUser'],
                    # 'Path': found['Name'],
                    'Contents': templates[0]['Value'],

                } for template in templates]
            }
        else:
            result = { "Revisions":
                [{
                    'RevisionId': str(template['Version']),
                    'Date': template['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                    'Key': key,
                    # 'Scope': found['Scope'],
                    # 'Origin': found['Origin'],
                    'User': template['LastModifiedUser'],
                    # 'Path': found['Name'],
                } for template in templates]
            }

        return result



    def delete(self, config, key):
        self.config = config
        Logger.debug(f"Locating SSM template '{key}': namespace:{config.namespace}, project={config.project}, application={config.application}, stage={config.stage}")
        ### only parameters owned by the context can be deleted, hence uninherited=True
        found = locate_ssm_parameter_in_context_hierachy(config=config, key=key, path_prefix=self.get_path_prefix(), uninherited=True)
        if not found:
            raise DSOException(f"Template '{key}' not found in the given context: namespace:{config.namespace}, project={config.project}, application={config.application}, stage={config.short_stage}")
        else:
            # if len(found) > 1:
            #     Logger.warn(f"More than one template found at '{found['Name']}'. The first one taken, and the rest were discarded.")
            if not found['Type'] == 'StringList':
                raise DSOException(f"Template '{key}' not found in the given context: namespace:{config.namespace}, project={config.project}, application={config.application}, stage={config.short_stage}")
        Logger.debug(f"Deleting SSM template: path={found['Name']}")
        delete_ssm_parameter(found['Name'])
        return {
                'Key': key, 
                'Stage': found['Stage'],
                'Scope': found['Scope'],
                'Origin': found['Origin'],
                'Path': found['Name'],
                }



def register():
    Providers.register(AwsSsmTemplateProvider())
